
#include "defs.h"

#include "dicts.h"
#include "eb123.h"
#include "ebook.h"
#include "icon.xpm"
#include "history.h"
#include "jcode.h"
#include "mainwnd.h"
#include "popupwnd.h"
#include "hotkeys.h"
#include "renderheadword.h"

static Mainwnd *_mainwnd = NULL;

G_DEFINE_TYPE(Mainwnd, mainwnd, GTK_TYPE_WINDOW);

static void mainwnd_class_init(MainwndClass *klass)
{
}

static void mainwnd_init(Mainwnd *self)
{
    if(!_mainwnd)
        _mainwnd = self;
    else
        g_assert(1);

    self->builder = g_object_new(TYPE_BUILDER, NULL);
    self->prefs = g_object_new(TYPE_PREFS, NULL);
    self->dicts = g_object_new(TYPE_DICTS, NULL);
    self->results = g_sequence_new(result_free);
    gtk_builder_connect_signals(GTK_BUILDER(self->builder), self);
    GtkWidget *vbox = GTK_WIDGET(gtk_builder_get_object(GTK_BUILDER(self->builder), "vbox1"));
    gtk_widget_reparent(vbox, GTK_WIDGET(self));
    GtkContainer *scroll = GTK_CONTAINER(gtk_builder_get_object(GTK_BUILDER(self->builder), "scrolledwindow2"));
    self->text = g_object_new(TYPE_TEXTVIEW, NULL);
    gtk_container_add(scroll, GTK_WIDGET(self->text));
    self->popupwnd = g_object_new(TYPE_POPUPWND, "mainwnd", self, "type", GTK_WINDOW_POPUP, NULL);
    g_printf("mainwd %.8X\n", self);
}

void mainwnd_headword_append(gpointer data, gpointer user_data)
{
    GtkListStore *store = GTK_LIST_STORE(user_data);
    GtkTreeIter iter;
    RESULT *res = (RESULT*)(data);
    gtk_list_store_append(store, &iter);
    gtk_list_store_set(store, &iter, 0, res->heading, 1, res, -1);
}

void mainwnd_update_results(Mainwnd *self)
{
    GtkListStore *store = GTK_LIST_STORE(gtk_builder_get_object(GTK_BUILDER(self->builder), "mainwnd_results_store"));
    gtk_list_store_clear(store);
    g_sequence_foreach(self->results, mainwnd_headword_append, store);
#if 1
    GtkTreeIter iter;
    GtkTreeView *tree = GTK_TREE_VIEW(gtk_builder_get_object(GTK_BUILDER(self->builder), "mainwnd_results"));
    GtkTreeSelection *select = gtk_tree_view_get_selection(tree);
    if(gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &iter))
    {
        gtk_tree_view_scroll_to_point(tree, 0, 0);
        gtk_tree_selection_select_iter(select, &iter);
	builder_grab_focus(self->builder, "mainwnd_results");
    }
#endif
}

static void mainwnd_results_results_selection_changed_cb(GtkTreeSelection *selection, gpointer data)
{
    Mainwnd *self = MAINWND(data);
    if(!selection)
	return;

    GtkTreeIter iter;
    GtkTreeModel *store;
    if(!gtk_tree_selection_get_selected(selection, &store, &iter))
        return;

    RESULT *res;
    //GtkTreeView *tree = GTK_TREE_VIEW(gtk_builder_get_object(GTK_BUILDER(self->builder), "mainwnd_results"));
    gtk_tree_model_get(store, &iter, 1, &res, -1);
    mainwnd_open(self, res);
    gtk_widget_show_all(GTK_WIDGET(self));
    gtk_window_present(GTK_WINDOW(self));
}

void mainwnd_exit()
{
    Mainwnd *mainwnd = _mainwnd;
    Prefs *prefs = mainwnd->prefs;

    GtkWidget *wnd = GTK_WIDGET(mainwnd);
    printf("save_prefs\n");
    int x, y, w, h;
    gdk_window_get_root_origin(wnd->window, &x, &y);
    w = wnd->allocation.width;
    h = wnd->allocation.height;
    prefs_set_int(prefs, "mainwnd.x", x);
    prefs_set_int(prefs, "mainwnd.y", y);
    prefs_set_int(prefs, "mainwnd.w", w);
    prefs_set_int(prefs, "mainwnd.h", h);

    GtkWidget *tree = GTK_WIDGET(gtk_builder_get_object(GTK_BUILDER(mainwnd->builder), "mainwnd_results"));
    w = tree->allocation.width;
    h = tree->allocation.height;
    prefs_set_int(prefs, "paned.tree_width", w);
    prefs_set_int(prefs, "paned.tree_height", h);

    //history_save_words();
    prefs_save(prefs);
    //dicts_save();
    //hotkeys_save();
    gtk_main_quit ();
}

void mainwnd_exit_btn_clicked_cb(GtkWidget *widget, gpointer data)
{
    mainwnd_exit();
}

void mainwnd_search_clear_clicked_cb(GtkWidget *widget, gpointer data)
{
    Mainwnd *wnd = MAINWND(data);
    GtkComboBox *combo = GTK_COMBO_BOX(gtk_builder_get_object(GTK_BUILDER(wnd->builder), "mainwnd_search"));
    GtkEntry *entry = GTK_ENTRY(gtk_bin_get_child(GTK_BIN(combo)));
    gtk_entry_set_text(entry, "");
    gtk_widget_grab_focus(GTK_WIDGET(entry));
}

void mainwnd_search(Mainwnd *self, const gchar *word)
{
    gint i = builder_get_int(self->builder, "mainwnd_search");
    gint method = ebook_search_method(i);
    //history_insert_word(word);

    gint maxhits = prefs_get_int(self->prefs, "headword.maxhits");
    g_printf("maxhits = %d\n", maxhits);
    ebook_search(word, method, self->results, maxhits, False);
    g_printf("results: %d\n", g_sequence_get_length(self->results));
    mainwnd_update_results(self);
    gtk_window_deiconify(GTK_WINDOW(self));
    gtk_widget_show_all(GTK_WIDGET(self));
    gtk_window_present(GTK_WINDOW(self));
}

void mainwnd_search_()
{
    gchar *word = builder_get_str(_mainwnd->builder, "mainwnd_search");
    mainwnd_search(_mainwnd, word);
}

#if 0
static void mainwnd_search_cb(GtkWidget *combo, gpointer user_data)
{
    mainwnd_search_();
}
#endif

void mainwnd_about_btn_clicked_cb(GtkWidget *w, gpointer data)
{
    gtk_show_about_dialog(GTK_WINDOW(data),
            "program-name", "eb123",
            "comments", _("A simple EPWING dictionary viewer"),
            "copyright", "Copyright \xc2\xa9 2009-2011 evgeny",
            "version", PACKAGE_VERSION,
            "website", "http://eb123.sourceforge.jp/",
            "website-label", "eb123.sourceforge.jp",
            NULL);
}

#if 0
static void mainwnd_prev(GtkWidget *widget, gpointer data)
{
    history_prev();
}
#endif

void mainwnd_clear_combo()
{
    builder_set_str(_mainwnd->builder, "mainwnd_search", "");
}

void mainwnd_clear_combo_cb(GtkWidget *widget, gpointer data)
{
    mainwnd_clear_combo();
}

gboolean mainwnd_key_press(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
    if(event->keyval == GDK_BackSpace)
    {
	GtkWidget *combo = GTK_WIDGET(gtk_builder_get_object(GTK_BUILDER(_mainwnd->builder), "mainwnd_search"));
	GtkEntry *entry = GTK_ENTRY(gtk_bin_get_child(GTK_BIN(combo)));
        if(!gtk_widget_has_focus(entry))
        {
            gtk_widget_grab_focus(mainwnd_search);
            gtk_editable_set_position(GTK_EDITABLE(entry), -1);
            return TRUE;
        }
    }
    return FALSE;
}

void mainwnd_reset_font()
{
    int custom_font = prefs_get_int(_mainwnd->prefs, "mainwnd.custom_font");
    gchar *fontstr = prefs_get_str(_mainwnd->prefs, "mainwnd.font");
    gchar *font = custom_font ? fontstr : "";
    gchar *s = g_strdup_printf("style \"custom-font\"{font_name=\"%s\"}\nclass \"GtkWidget\" style \"custom-font\"", font);
    gtk_rc_parse_string(s);
    g_free(s);
    GdkScreen *screen = gtk_window_get_screen(mainwnd_get_wnd());
    GtkSettings *settings = gtk_settings_get_for_screen(screen);
    gtk_rc_reset_styles(settings);
}

void mainwnd_status_icon_menu(GtkStatusIcon *status_icon, guint button, guint activate_time, gpointer data)
{
    GtkWidget *menu = gtk_menu_new(), *item;
    popupwnd.scan_status_icon_menu = gtk_check_menu_item_new_with_label(_("Scan"));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), popupwnd.scan_status_icon_menu);
    //g_signal_connect(G_OBJECT(popupwnd.scan_status_icon_menu), "toggled", G_CALLBACK(selection_toggled_status_icon_menu), NULL);
    //gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(popupwnd.scan_status_icon_menu), selection.lookup_started);
    item = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, NULL);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
    g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(mainwnd_exit_btn_clicked_cb), NULL);
    gtk_widget_show_all(menu);
    gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 0, gtk_get_current_event_time());
}

void mainwnd_iconify_restore()
{
    if(gtk_widget_get_visible(GTK_WIDGET(mainwnd_get_wnd())))
    {
	prefs_close_wnd();
	gtk_widget_hide(GTK_WIDGET(mainwnd_get_wnd()));
    }
    else
	gtk_window_present(mainwnd_get_wnd());
}

void mainwnd_iconify_cb(GtkWidget *widget, gpointer data)
{
    mainwnd_iconify_restore();
}

void mainwnd_status_icon_add()
{
    GdkPixbuf *pixbuf;
    GtkStatusIcon *icon;
    pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)icon_xpm);
    icon = gtk_status_icon_new_from_pixbuf(pixbuf);
    g_signal_connect(G_OBJECT(icon), "activate", G_CALLBACK(mainwnd_iconify_cb), NULL);
    g_signal_connect(G_OBJECT(icon), "popup_menu", G_CALLBACK(mainwnd_status_icon_menu), NULL);
    g_object_unref(pixbuf);
}

void mainwnd_dict_groups_update(Mainwnd *self)
{
    GtkTreeStore *dicts_store = GTK_TREE_STORE(gtk_builder_get_object(GTK_BUILDER(self->builder), "dicts_store"));
    GtkListStore *groups = GTK_LIST_STORE(gtk_builder_get_object(GTK_BUILDER(self->builder), "mainwnd_dictgroups_store"));
    gtk_list_store_clear(groups);
    GtkTreeIter iter1, iter2;
    if(gtk_tree_model_get_iter_first(GTK_TREE_MODEL(dicts_store), &iter1))
    {
	do { 
	    gchar *title;
	    gtk_tree_model_get(GTK_TREE_MODEL(dicts_store), &iter1, DICT_ALIAS, &title, -1);
	    gtk_list_store_append(groups, &iter2);
	    gtk_list_store_set(groups, &iter2, 0, title, -1);
	    g_free(title);
	} while(gtk_tree_model_iter_next(GTK_TREE_MODEL(dicts_store), &iter1));
    }
    builder_set_int(self->builder, "mainwnd_dict_group", 0);
}

void mainwnd_dict_buttons_remove(Mainwnd *self)
{

}

void mainwnd_dict_toggled_cb(GtkWidget *w, gpointer data)
{
    BOOK_INFO *binfo = (BOOK_INFO*)data;
    binfo->active = gtk_toggle_tool_button_get_active(GTK_TOGGLE_TOOL_BUTTON(w));
}

void mainwnd_find_btn_clicked_cb(GtkWidget *w, gpointer data)
{
    Mainwnd *mainwnd = MAINWND(data);
    gchar *word = builder_get_str(mainwnd->builder, "mainwnd_search");
    g_printf("word: %s\n", word);
    mainwnd_search(mainwnd, word);
}

void mainwnd_dict_buttons_add(Mainwnd *self)
{
    gint n = builder_get_int(self->builder, "mainwnd_dict_group");
    GtkTreeIter iter;
    if(!dicts_get_nth(self->dicts, n, &iter))
	return;
    GtkToolbar *tb = GTK_TOOLBAR(gtk_builder_get_object(GTK_BUILDER(self->builder), "mainwnd_dicts"));
    gchar *title;
    BOOK_INFO *binfo;
    GtkTreeModel *store = GTK_TREE_MODEL(self->dicts->store);
    do
    {
	gtk_tree_model_get(store, &iter, DICT_ALIAS, &title, DICT_BINFO, &binfo, -1);
	GtkToolItem *toggle = gtk_toggle_tool_button_new();
        gtk_tool_button_set_label(GTK_TOOL_BUTTON(toggle), title);
        gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(toggle), binfo->active);
        gtk_tool_item_set_homogeneous(toggle, FALSE);
        g_signal_connect(G_OBJECT(toggle),"toggled", G_CALLBACK(mainwnd_dict_toggled_cb), (gpointer)binfo);

	gtk_toolbar_insert(tb, toggle, -1);
	if(!binfo->available)
            gtk_widget_set_sensitive(GTK_WIDGET(toggle), FALSE);
	gtk_widget_set_tooltip_text(GTK_WIDGET(toggle), title);
	g_free(title);
	if(!gtk_tree_model_iter_next(store, &iter))
            break;
        gtk_toolbar_insert(tb, gtk_separator_tool_item_new(), -1);
    } while(TRUE);
}

void mainwnd_dict_buttons_update(Mainwnd *self)
{
    mainwnd_dict_buttons_remove(self);
    mainwnd_dict_buttons_add(self);
}

void mainwnd_prepare(Mainwnd *self)
{
#if 0
    hotkeys_local_install();
    hotkeys_global_install();
#endif
    builder_install_text_cellrenderer(self->builder, "mainwnd_search_method");
    builder_install_text_cellrenderer(self->builder, "mainwnd_dict_group");
    GtkTreeView *tree = GTK_TREE_VIEW(gtk_builder_get_object(GTK_BUILDER(self->builder), "mainwnd_results"));
    GtkTreeSelection *select = gtk_tree_view_get_selection(tree);
    gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);
    GtkCellRenderer *renderer = g_object_new(TYPE_RENDER_HEADWORD, NULL); //gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(tree, -1, _("Result"), renderer, "text", 0, "result", 1, NULL);
    gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);
    g_signal_connect(G_OBJECT(select), "changed", G_CALLBACK(mainwnd_results_results_selection_changed_cb), self);

    GtkComboBox *combo = GTK_COMBO_BOX(gtk_builder_get_object(GTK_BUILDER(self->builder), "mainwnd_search"));
    g_signal_connect(gtk_bin_get_child(GTK_BIN(combo)), "activate", G_CALLBACK(mainwnd_find_btn_clicked_cb), self);
    
    gint active = dicts_load(self->dicts);
    mainwnd_dict_groups_update(self);
    mainwnd_dict_buttons_update(self);
    builder_set_int(self->builder, "mainwnd_dict_group", active);
    builder_set_int(self->builder, "mainwnd_search", 0);

    gint remember_pos = prefs_get_int(self->prefs, "mainwnd.remember_pos");
    if(remember_pos)
    {
	gint x = prefs_get_int(self->prefs, "mainwnd.x");
	gint y = prefs_get_int(self->prefs, "mainwnd.y");
	gint w = prefs_get_int(self->prefs, "mainwnd.w");
	gint h = prefs_get_int(self->prefs, "mainwnd.h");
	gtk_window_set_default_size(GTK_WINDOW(self), w, h);
        gtk_window_move(GTK_WINDOW(self), x, y);
	w = prefs_get_int(self->prefs, "paned.tree_width");
	h = prefs_get_int(self->prefs, "paned.tree_height");
	gtk_widget_set_size_request(tree, w, h);
    }
    builder_grab_focus(self->builder, "mainwnd_search");
}

void mainwnd_dicts_update()
{
    mainwnd_dict_groups_update(_mainwnd);
    mainwnd_dict_buttons_update(_mainwnd);
}

gboolean mainwnd_get_dicts(GtkTreeModel **store, GtkTreeIter *iter)
{
    Mainwnd *self = _mainwnd;
    int n = builder_get_int(self->builder, "mainwnd_dict_group");
    *store = GTK_TREE_MODEL(gtk_builder_get_object(GTK_BUILDER(self->builder), "dicts_store"));
    GtkTreeIter iter1;
    if(!gtk_tree_model_iter_nth_child(*store, &iter1, NULL, n))
	return FALSE;
    if(!gtk_tree_model_iter_children(*store, iter, &iter1))
        return FALSE;
    return TRUE;
}

GtkWindow *mainwnd_get_wnd()
{
    return GTK_WINDOW(_mainwnd);
}

GSequence *mainwnd_get_results()
{
    return _mainwnd->results;
}

void mainwnd_insert_text(gchar *txt, gboolean clear)
{
    TextView *text = _mainwnd->text;
    textview_insert_message(text, txt, clear);
}

void mainwnd_clear_text()
{
    g_printf("TODO: clear text\n");
}

gint mainwnd_get_search_method()
{
    return builder_get_int(_mainwnd->builder, "mainwnd_search_method");
}

void mainwnd_open(Mainwnd *self, RESULT *res)
{
    textview_open(self->text, res, self->prefs, TRUE, 0);
    gtk_widget_show_all(GTK_WIDGET(mainwnd_get_wnd()));
}

void mainwnd_scan_toggled_cb(GtkToggleButton *btn, gpointer data)
{
    Mainwnd *w = MAINWND(data);
    if(gtk_toggle_button_get_active(btn))
    {
	popupwnd_lookup_start(w->popupwnd);
	//selection_search_in_popupwnd();
    }
    else
	popupwnd_lookup_stop(w->popupwnd);
}

