
#include "defs.h"

#include "dicts.h"
#include "ebook.h"
#include "jcode.h"
#include "history.h"
#include "mainwnd.h"
#include "popupwnd.h"
#include "textview.h"

static Popupwnd *_popupwnd = NULL;

#define SELECTION_TIMEOUT 300

G_DEFINE_TYPE(Popupwnd, popupwnd, GTK_TYPE_WINDOW);

static void popupwnd_set_property(GObject *object, guint param_id, const GValue *value, GParamSpec *pspec)
{
    Popupwnd *w = POPUPWND(object);

    switch(param_id)
    {
	case 1:
	    w->mainwnd = g_value_get_pointer(value);
	    Mainwnd *mainwnd = MAINWND(w->mainwnd);
	    w->prefs = mainwnd->prefs;
	    w->builder = mainwnd->builder;
	    GtkWidget *vbox = GTK_WIDGET(gtk_builder_get_object(GTK_BUILDER(w->builder), "vbox9"));
	    gtk_widget_reparent(vbox, GTK_WIDGET(w));
	    GtkContainer *scroll = GTK_CONTAINER(gtk_builder_get_object(GTK_BUILDER(w->builder), "popupwnd_scroll"));
	    w->text = g_object_new(TYPE_TEXTVIEW, NULL);
	    gtk_container_add(scroll, GTK_WIDGET(w->text));
	    break;
	default:
	    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, param_id, pspec);
	    break;
    }
}

static void popupwnd_get_property(GObject *object, guint param_id, GValue *value, GParamSpec *pspec)
{
    Popupwnd *w = POPUPWND(object);
    switch(param_id)
    {
	case 1:
	    g_value_set_pointer(value, w->mainwnd);
	    break;
	default:
	    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
	    break;
    }
}

static void popupwnd_class_init(PopupwndClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->set_property = popupwnd_set_property;
    gobject_class->get_property = popupwnd_get_property;
    g_object_class_install_property(gobject_class, 1, g_param_spec_pointer("mainwnd", _("Mainwnd"), _("Mainwnd"), G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
}

static void popupwnd_init(Popupwnd *self)
{
    if(!_popupwnd)
	_popupwnd = self;
    else
	g_assert(1);
    //Mainwnd *mainwnd = MAINWND(self->mainwnd);
    //g_printf("popupwnd constructor mainwnd %.8X\n", self->mainwnd);
    self->timeout_id = 0;
    self->lookup_started = False;
    self->lookup_suspended = True;
    self->drag = False;
    self->menu = NULL;
    self->binfo = NULL;
    self->results = g_sequence_new(result_free);
    self->x = -1;
    self->y = -1;
    gtk_window_set_position(GTK_WINDOW(self), GTK_WIN_POS_CENTER);
}

static const gchar* popupwnd_selection_copy(Popupwnd *self, gboolean force_copy)
{
    GtkClipboard* clipboard = gtk_clipboard_get(GDK_SELECTION_PRIMARY);
    gchar *str = gtk_clipboard_wait_for_text(clipboard);
    static gchar *sel = NULL, *sel_prev = NULL;
    gint maxchar = prefs_get_int(self->prefs, "selection.maxchar");
    if(!str || (self->lookup_suspended && !force_copy))
    {
        if(sel_prev)
            g_free(sel_prev);
        sel_prev = NULL;
        return NULL;
    }
    g_strstrip(str);
    if(strlen(str) > maxchar)
    {
        if(sel_prev)
            g_free(sel_prev);
        sel_prev = NULL;
        g_free(str);
        return NULL;
    }
    if(force_copy)
        sel = str;
    else
    {
        if(!sel)
        {
            sel = str;
            return NULL;
        }
        if(strncmp(str, sel, maxchar))
        {
            g_free(sel);
            sel = str;
            return NULL;
        }
        g_free(str);
        if((sel_prev) ? !strncmp(sel_prev, sel, maxchar) : FALSE)
            return NULL;
    }
    sel_prev = sel;
    sel = NULL;
    return sel_prev;
}

static void popupwnd_dict_changed_cb(GtkWidget *w, gpointer data)
{
    _popupwnd->binfo = (BOOK_INFO*)data;
    //_popupwnd->lock = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(popupwnd.lockbtn));
    //popupwnd_search(NULL);
}

static gboolean popupwnd_dict_menu_append(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
    Popupwnd *w = POPUPWND(data);
    static GSList *group = NULL;
    static GtkWidget *submenu = NULL;
    GtkWidget *item;
    if(!model)
    {
        if(w->menu)
            gtk_widget_destroy(w->menu);
        group = NULL;
	w->menu = gtk_menu_new();
        item = gtk_radio_menu_item_new_with_label(group, _("Main window selection"));
        g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(popupwnd_dict_changed_cb), NULL);
        group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(item));
        gtk_menu_shell_append(GTK_MENU_SHELL(w->menu), item);
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), TRUE); 
        submenu = NULL;
        return FALSE;
    }
    if(gtk_tree_path_get_depth(path) == 1)
    {
        gchar *name;
        submenu = gtk_menu_new();
        gtk_tree_model_get(model, iter, DICT_ALIAS, &name, -1);
        gtk_menu_shell_append(GTK_MENU_SHELL(w->menu), item = gtk_menu_item_new_with_label(name));
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), submenu);
        g_free(name);
    }
    else
    {
        gchar *name;
        BOOK_INFO *binfo;
        gtk_tree_model_get(model, iter, DICT_ALIAS, &name, DICT_BINFO, &binfo, -1);
        item = gtk_radio_menu_item_new_with_label(group, binfo->title);
        group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(item));
        gtk_menu_shell_append(GTK_MENU_SHELL(submenu), item);
	if(binfo == w->binfo)
            gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), TRUE); 
        g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(popupwnd_dict_changed_cb), binfo);
        g_free(name);
        
    }
    return FALSE;
}

static void popupwnd_dict_menu_create(Popupwnd *self, gboolean newmenu)
{
    if(newmenu || !popupwnd.menu)
    {
        popupwnd_dict_menu_append(NULL, NULL, NULL, self);
	GtkTreeModel *store = GTK_TREE_MODEL(gtk_builder_get_object(GTK_BUILDER(self->builder), "dicts_store"));
        gtk_tree_model_foreach(GTK_TREE_MODEL(store), popupwnd_dict_menu_append, self);
	if(self->binfo)
	{
	    if(!dicts_find_path(self->binfo->book))
		self->binfo = NULL;
	}
    }
}

void popupwnd_dict_btn_clicked_cb(GtkWidget *widget, gpointer data)
{
    popupwnd_dict_menu_create(_popupwnd, False);
    _popupwnd->lock = TRUE;
    gtk_widget_show_all(_popupwnd->menu);
    gtk_menu_popup(GTK_MENU(_popupwnd->menu), NULL, NULL, NULL, NULL, 0, gtk_get_current_event_time());
}

void popupwnd_render(gpointer data, gpointer user_data)
{
    Popupwnd *w = POPUPWND(user_data);
    RESULT *res = (RESULT*)data;
    textview_open(w->text, res, w->prefs, FALSE, 0);
}

void popupwnd_search(Popupwnd *self, const gchar *str)
{
    g_printf("search: %s\n", str);

    gint search_method = prefs_get_int(self->prefs, "popupwnd.search_method");
    gint maxhits = prefs_get_int(self->prefs, "popupwnd.maxhits");
    gint jap_only = prefs_get_int(self->prefs, "popupwnd.jap_only");
    gboolean b = False;
    if(self->binfo)
        b = ebook_search_book(str, popupwnd.search_method, self->results, maxhits, self->binfo, jap_only);
    else
        b = ebook_search(str, popupwnd.search_method, self->results, maxhits, jap_only);
    if(!b)
	return;

    popupwnd_dict_menu_create(self, False);
    GtkLabel *label = GTK_LABEL(gtk_builder_get_object(GTK_BUILDER(self->builder), "popupwnd_label"));
    gtk_label_set_text(label, str);

    textview_clear_textbuf(self->text);
    if(g_sequence_get_length(self->results))
	g_sequence_foreach(self->results, popupwnd_render, self);
    else
	textview_insert_message(self->text, "<No hits>", TRUE);
    gtk_window_present(GTK_WINDOW(self));
    if((self->x > 0) && (self->y > 0))
	gtk_window_move(GTK_WINDOW(_popupwnd), self->x, self->y);

    gint w = prefs_get_int(self->prefs, "popupwnd.w");
    gint h = prefs_get_int(self->prefs, "popupwnd.h");
    w = (w > 100) ? w : 100;
    h = (h > 100) ? h : 100;
    gtk_widget_set_size_request(GTK_WIDGET(self), w, h);

    gtk_widget_show_all(GTK_WIDGET(self));
}

gint popupwnd_search_cb(gpointer data)
{
    Popupwnd *self = POPUPWND(data);
    const gchar *sel = popupwnd_selection_copy(self, False);
    if(sel)
        popupwnd_search(self, sel);
    return 1;
}

void popupwnd_lookup_resume(Popupwnd *self)
{
    if(self->lookup_started)
    {
	if(self->timeout_id != 0)
	    g_source_remove(self->timeout_id);
	self->timeout_id = g_timeout_add(SELECTION_TIMEOUT, popupwnd_search_cb, self);
	//gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(popupwnd.scan), TRUE);
    }
    self->lookup_suspended = FALSE;
}

void popupwnd_lookup_start(Popupwnd *self)
{
    self->lookup_started = TRUE;
    //gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(popupwnd.scan), TRUE);
    popupwnd_lookup_resume(self);
}

void popupwnd_lookup_suspend(Popupwnd *self)
{
    if(self->timeout_id != 0)
	g_source_remove(self->timeout_id);
    self->timeout_id = 0;
    //popupwnd_close(NULL, NULL);
    self->lookup_suspended = TRUE;
}

void popupwnd_lookup_stop(Popupwnd *self)
{
    self->lookup_started = False;
    //gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(popupwnd.scan), TRUE);
    popupwnd_lookup_suspend(self);
}

void popupwnd_close(GtkWidget *widget, gpointer data)
{
    if(popupwnd.wnd)
    {
	gtk_window_get_position(GTK_WINDOW(popupwnd.wnd), &popupwnd.x, &popupwnd.y);
        gtk_widget_destroy(popupwnd.wnd);
        popupwnd.wnd = NULL;
    }
}

static gboolean popupwnd_close_(gpointer data)
{
    popupwnd_close(NULL, NULL);
    return FALSE;
}

static void popupwnd_start_stop_timer(gboolean start)
{
    static gint timeout_id = 0;
    if(timeout_id != 0)
        g_source_remove(timeout_id);
    timeout_id = 0;
    if(!start || popupwnd.lock)
        return;
    timeout_id = g_timeout_add(popupwnd.timeout, popupwnd_close_, NULL);
}

void popupwnd_leave_notify(GtkWidget *widget, GdkEventCrossing *event, gpointer user_data)
{
    popupwnd_start_stop_timer((event->detail == GDK_NOTIFY_NONLINEAR) || (event->detail == GDK_NOTIFY_NONLINEAR_VIRTUAL));
}

void popupwnd_enter_notify(GtkWidget *widget, GdkEventCrossing *event, gpointer user_data)
{
    popupwnd_start_stop_timer(FALSE);
}

void popupwnd_lock(GtkWidget *widget, gpointer user_data)
{
    popupwnd.lock = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
}

void popupwnd_query_in_mainwnd(GtkWidget *widget, gpointer data)
{
    const gchar *str = gtk_label_get_text(GTK_LABEL(popupwnd.label));
    //mainwnd_search(str);
    gtk_window_deiconify(mainwnd_get_wnd());
}

void popupwnd_close_btn_clicked_cb(GtkWidget *widget, gpointer data)
{
    //Mainwnd *w = MAINWND(data);
    //GtkWindow *w = gtk_widget_get_parent_window(widget);
    gtk_widget_hide(GTK_WIDGET(_popupwnd));
}

gint popupwnd_evtbox_motion_notify_event_cb(GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
    if(!_popupwnd->drag)
	return TRUE;
    gint x = _popupwnd->dx + event->x_root;
    gint y = _popupwnd->dy + event->y_root;
    if(x < 0) x = 0;
    if(y < 0) y = 0;
    gtk_window_move(GTK_WINDOW(_popupwnd), x, y);
    _popupwnd->x = x;
    _popupwnd->y = y;
    return TRUE;
}

gint popupwnd_evtbox_button_press_event_cb(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
    if(event->type==GDK_2BUTTON_PRESS)
    {
	_popupwnd->drag = FALSE;
	//popupwnd_close(NULL, NULL);
    }
    else
    {
	_popupwnd->drag = TRUE;
	gint ex, ey, wx, wy;
	ex = event->x_root;
	ey = event->y_root;
	gtk_window_get_position(GTK_WINDOW(_popupwnd), &wx, &wy);
	_popupwnd->dx = wx - ex;
	_popupwnd->dy = wy - ey;
    }
    return FALSE;
}

gint popupwnd_evtbox_button_release_event_cb(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
    _popupwnd->drag = FALSE;
    return FALSE;
}

static void popupwnd_scroll_mark(GtkTextBuffer *buffer, gboolean down)
{
    gint i, x, y, n, len, d;;
    gchar buf[16];
    GtkTextIter iter, iter1;
    GtkTextMark *mark;
    gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(popupwnd.view), GTK_TEXT_WINDOW_TEXT, 1, 1, &x, &y);
    gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(popupwnd.view), &iter, x, y);
    len = g_sequence_get_length(popupwnd.results);
    if(down)
    {
        n = 0; d = 1;
    }
    else
    {
        n = len - 1; d = -1;
    }
    for(i = n;; i += d)
    {
        sprintf(buf, "%d", i);
        mark = gtk_text_buffer_get_mark(buffer, buf);
        if(!mark)
            break;
        gtk_text_buffer_get_iter_at_mark(buffer, &iter1, mark);
        if(gtk_text_iter_compare(&iter, &iter1) == -d)
        {
            gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(popupwnd.view), mark, 0.0, TRUE, 0.0, 0.0);
            break;
        }
    }
}

static void popupwnd_scroll_down_cb(GtkWidget *widget, gpointer data)
{
    popupwnd_scroll_mark(gtk_text_view_get_buffer(GTK_TEXT_VIEW(popupwnd.view)), TRUE);
}

static void popupwnd_scroll_up_cb(GtkWidget *widget, gpointer data)
{
    popupwnd_scroll_mark(gtk_text_view_get_buffer(GTK_TEXT_VIEW(popupwnd.view)), FALSE);
}

#if 0
static void popupwnd_window_create()
{
    GdkModifierType mask;
    gint pos_x, pos_y;
    gint pointer_x, pointer_y;
    gint root_x, root_y;

    GtkWidget *vbox, *hbox, *eventbox, *frame, *btn, *image, *scroll;

    if(popupwnd.remember_pos && (popupwnd.x != -1) && (popupwnd.y != -1))
    {
	pos_x = popupwnd.x;
	pos_y = popupwnd.y;
    }
    else
    {
	GdkWindow *root_win = gdk_x11_window_foreign_new_for_display(gdk_display_get_default(), GDK_ROOT_WINDOW());
	root_x = gdk_window_get_width(root_win);
	root_y = gdk_window_get_height(root_win);
	gdk_window_get_pointer(root_win, &pointer_x, &pointer_y, &mask);
	pos_x = pointer_x + 10;
	pos_y = pointer_y + 10;

	if(pos_x + popupwnd.w > root_x)
	    pos_x = root_x - popupwnd.w;
	
	if(pos_y + popupwnd.h > root_y)
	    pos_y = root_y - popupwnd.h;
    }

    popupwnd.wnd = gtk_widget_new(GTK_TYPE_WINDOW,
	    "type", GTK_WINDOW_POPUP,
	    "default-width", popupwnd.w,
	    "default-height", popupwnd.h,
	    NULL);
    gtk_window_move(GTK_WINDOW(popupwnd.wnd), pos_x, pos_y);
    g_signal_connect(G_OBJECT(popupwnd.wnd), "delete_event", G_CALLBACK(popupwnd_close), NULL);
    g_signal_connect(G_OBJECT(popupwnd.wnd), "leave_notify_event", G_CALLBACK(popupwnd_leave_notify), (gpointer)NULL);
    g_signal_connect(G_OBJECT(popupwnd.wnd), "enter_notify_event", G_CALLBACK(popupwnd_enter_notify), (gpointer)NULL);

    vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(popupwnd.wnd), vbox);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
    gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    eventbox = gtk_event_box_new();
    gtk_box_pack_start(GTK_BOX(hbox), eventbox, TRUE, TRUE, 2);
    g_signal_connect(G_OBJECT(eventbox), "button_press_event", G_CALLBACK(popupwnd_title_button_press_cb), (gpointer)NULL);
    g_signal_connect(G_OBJECT(eventbox), "motion_notify_event", G_CALLBACK(popupwnd_title_motion_notify_cb), (gpointer)NULL);
    g_signal_connect(G_OBJECT(eventbox), "button_release_event", G_CALLBACK(popupwnd_title_button_release_cb), (gpointer)NULL);
    popupwnd.label = gtk_label_new("");
    gtk_container_add(GTK_CONTAINER(eventbox), popupwnd.label);

    popupwnd.lockbtn = gtk_toggle_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(popupwnd.lockbtn), popupwnd.lock);
    gtk_button_set_relief(GTK_BUTTON(popupwnd.lockbtn), GTK_RELIEF_NONE);
    g_signal_connect(G_OBJECT(popupwnd.lockbtn), "toggled", G_CALLBACK(popupwnd_lock), (gpointer)NULL);
    gtk_box_pack_start(GTK_BOX(hbox), popupwnd.lockbtn, FALSE, FALSE, 0);
    gtk_widget_set_tooltip_text(popupwnd.lockbtn, _("Lock popup window on the screen."));
    image = gtk_image_new_from_stock(GTK_STOCK_REFRESH, GTK_ICON_SIZE_SMALL_TOOLBAR);
    gtk_container_add(GTK_CONTAINER(popupwnd.lockbtn), image);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
    gtk_box_pack_start(GTK_BOX(vbox), frame, TRUE, TRUE, 0);

    scroll = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_container_add(GTK_CONTAINER(frame), scroll);

    popupwnd.view = g_object_new(TYPE_TEXTVIEW, NULL);
    gtk_container_add(GTK_CONTAINER(scroll), GTK_WIDGET(popupwnd.view));

    gtk_widget_realize(popupwnd.wnd);
    gdk_window_set_decorations(popupwnd.wnd->window, 0);

    gtk_widget_show_all(popupwnd.wnd);
}
#endif

static gint popupwnd_scroll_to_top()
{
    GtkTextIter iter;
    GtkTextMark *mark;
    GtkTextBuffer *buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(popupwnd.view));
    gtk_text_buffer_get_start_iter(buf, &iter);
    mark =  gtk_text_buffer_create_mark(buf, "start", &iter, TRUE);
    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(popupwnd.view), mark, 0.0, TRUE, 0.0, 0.0);
    gtk_text_buffer_delete_mark(buf, mark);
    return 0;
}

void popupwnd_open(RESULT *res)
{
    static int counter = 0;
    if(!res)
    {
        counter = 0;
        return;
    }
    //textview_open(popupwnd.view, res, FALSE, counter++);
    popupwnd_scroll_to_top();
}

void popupwnd_show(const gchar *word)
{
    textview_clear_textbuf(popupwnd.view);
    popupwnd_open(NULL);
    if(g_sequence_get_length(popupwnd.results))
        g_sequence_foreach(popupwnd.results, popupwnd_render, NULL);
    else
        textview_insert_message(popupwnd.view, "No hit.", TRUE);
    gtk_label_set_text(GTK_LABEL(popupwnd.label), word);
    gtk_window_present(GTK_WINDOW(popupwnd.wnd));
}

#if 0
void popupwnd_search(const gchar *str)
{
    if(!str)
        str = gtk_label_get_text(GTK_LABEL(popupwnd.label));

    gchar *euc_str = iconv_convert(ENC_UTF8, ENC_EUC_JP, str);
    if(!popupwnd_validate_euc(euc_str))
    {
        g_free(euc_str);
        return;
    }
    if(!popupwnd.wnd)
    {
        popupwnd_window_create();
	if(!popupwnd.menu)
	    popupwnd.binfo = NULL;
    }
    popupwnd_dict_menu_create(False);
    if(popupwnd.binfo)
        ebook_search_book(euc_str, popupwnd.search_method, popupwnd.results, popupwnd.maxhits, popupwnd.binfo);
    else
        ebook_search(euc_str, popupwnd.search_method, popupwnd.results, popupwnd.maxhits);
    g_free(euc_str);
    popupwnd_show(str);
}
#endif
